import uvicorn
import argparse
from semeval.serviecs import embeddings

if __name__ == '__main__':
    p = argparse.ArgumentParser()
    p.add_argument("--host", default=None, help="Hostname (default: localhost)")
    p.add_argument("--port", default=None, help="Port (default: 1337)")
    p.add_argument("--service", required=True, help="The service to run (e.g. embeddings)")
    args = p.parse_args()

    services = {
        'embeddings': embeddings.EmbeddingsServer,
    }

    if args.service not in services:
        raise Exception("Service '{}' not available!".format(args.service))

    uvicorn.run(services[args.service](**args.__dict__),
                host=args.host if args.host else "localhost",
                port=int(args.port) if args.port else 1337)



from gensim.models import KeyedVectors
import numpy as np
from pathlib import Path
from .common import *


class Embeddings(object):
    def __init__(self, lang='eng'):
        model_path = Path(download_path() + "vectors-{}.txt".format(lang))
        if lang not in supported_languages():
            raise Exception("Language '{}' is not supported!".format(lang))
        elif not model_path.is_file():
            raise Exception("Vectors for language '{}' are not downloaded! "
                            "Download them using `python -m semeval.download -m {} -l {}`"
                            .format(lang, 'embeddings', lang))

        self.L = KeyedVectors.load_word2vec_format(model_path, binary=False, unicode_errors='replace')
        self.L_len = len(self.L.vocab)

    @staticmethod
    def align(e1, e2, word, topn=10):
        return e2.L.similar_by_vector(e1.L.word_vec(word), topn=topn)

    def vector(self, word):
        return self.L[word]

    def project(self, word, e2, topn=10):
        return self.align(self, e2, word, topn=topn)

    def similarity(self, *args, **kwargs):
        return self.L.similarity(*args, **kwargs)

    def most_similar(self, *args, **kwargs):
        return self.L.most_similar_cosmul(*args, **kwargs)

    def theme(self, l):
        return self.L.most_similar(positive=l, topn=1)[0]

    def vocabulary(self):
        return self.L.vocab

    def centroid(self, l):
        vectors = [self.L.get_vector(w) for w in l if w in self.L.vocab]
        return np.mean(vectors, axis=0)

    def neighbours(self, w, topn=50):
        return self.L.most_similar(w, topn=topn)

    def neighbours_threshold(self, w, threshold=0.8):
        return [x for x in self.L.most_similar(w, topn=self.L_len) if x[1] >= threshold]

    def analogy(self, a, b, c, topn=10):
        return list(self.L.most_similar(positive=[b, c], negative=[a], topn=topn))

    def to_vector(self, tokens):
        text_v = [self.L.get_vector(t) for t in tokens if t in self.L.vocab]

        if len(text_v) == 0:
            raise Exception("No words in the text found in the model.")
        return np.mean(text_v, axis=0)



import argparse
from .common import *

ZENODO_BASEURL = "https://zenodo.org/record/5156189/files/"


def _file_name(lang, model):
    if model == "sentiment":
        return "sentiment_model.pt"
    elif model == "embeddings":
        if lang not in supported_languages():
            raise (BaseException("Language not supported!"))
        return "vectors-{}.txt".format(lang)
    else:
        raise (BaseException("Unknown model type " + model))


def download_relatedness_model(language):
    relatedness_models = {
        'eng': {
            'rows': 'https://zenodo.org/record/4633293/files/rows.txt',
            'cols': 'https://zenodo.org/record/4633293/files/cols.txt',
            'model': 'https://zenodo.org/record/4633293/files/model_gzip.hkl'
        },
        'fin': {
            'rows': 'https://zenodo.org/record/3473456/files/unigrams_sorted_5k.txt',
            'cols': 'https://zenodo.org/record/3473456/files/unigrams_sorted_5k.txt',
            'model': 'https://zenodo.org/record/3473456/files/rel_matrix_n_csr.hkl'
        }
    }
    if language not in relatedness_models:
        raise (BaseException("Language not supported!"))

    for _f_name, _link in relatedness_models[language].items():
        _ext = _link.split('.')[-1]
        download_file("{}?download=1".format(_link),
                      "{}{}-relatedness-{}.{}".format(download_path(), language, _f_name, _ext),
                      show_progress=True)


def main(languages, models):
    os.makedirs(script_path("data"), exist_ok=True)
    for language in languages:
        for model in models:
            if model == 'relatedness':
                download_relatedness_model(language)
            else:
                file = _file_name(language, model)
                print("Downloading", model, "for", language)
                download_file(ZENODO_BASEURL + file + "?download=1", download_path() + file,
                              show_progress=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='semeval download models')
    parser.add_argument('-l', '--languages', nargs='+', help='<Required> languages to download', required=True)
    parser.add_argument('-m', '--models', nargs='+', help='<Required> models to download', required=True)
    args = parser.parse_args()

    main(args.languages, args.models)



from typing import Optional, List
from fastapi import FastAPI, Query, Request, HTTPException
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
import ujson as json
from semeval.embeddings import Embeddings
import json
import requests
from semeval.common import *


def EmbeddingsServer(*args, **kwargs):
    models = {}
    app = FastAPI()

    for lang in kwargs.get('languages', []):
        models[lang] = Embeddings(lang)

    def filter_words(lang, words):
        if words is None:
            return
        return [word for word in words if word in models[lang].L.vocab]

    def check_language(lang):
        if lang not in models and lang in supported_languages():
            models[lang] = Embeddings(lang)

    @app.middleware("http")
    async def load_language(request: Request, call_next):
        try:
            for _lang_k in ['lang', 'lang1', 'lang2']:
                if _lang_k in request.query_params:
                    check_language(request.query_params[_lang_k])
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

        response = await call_next(request)
        return response

    @app.get("/n_similarity/")
    def n_similarity(lang: str, ws1: List[str] = Query([]), ws2: List[str] = Query([])):
        try:
            ws1 = filter_words(lang, ws1)
            ws2 = filter_words(lang, ws2)
            return models[lang].L.n_similarity(ws1, ws2)
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/similarity/")
    def similarity(lang: str, w1: str, w2: str):
        try:
            res = {'w1': w1, 'w2': w2, 'score': models[lang].L.similarity(w1, w2).item()}
            return JSONResponse(json.dumps(res))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/most_similar/")
    def most_similar(lang: str, positive: List[str] = Query([]), negative: List[str] = Query([]), topn: int = 10):
        try:
            positive = filter_words(lang, positive)
            negative = filter_words(lang, negative)
            res = models[lang].L.most_similar_cosmul(positive=positive, negative=negative, topn=topn)
            return JSONResponse(json.dumps(res))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/neighbours/")
    def neighbours(lang: str, word: str, threshold: float = 0.8):
        try:
            return JSONResponse(models[lang].neighbours_threshold(word, threshold))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/model_word_set/")
    def model_word_set(lang: str):
        try:
            res = models[lang].L.index2word
            return JSONResponse(json.dumps(res))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/to_vector/")
    def to_vector(lang: str, tokens: List[str] = Query([])):
        try:
            res = models[lang].to_vector(tokens)
            return JSONResponse(json.dumps(res.tolist()))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/theme/")
    def theme(lang: str, words: List[str] = Query([])):
        try:
            words = filter_words(lang, words)
            res = models[lang].theme(words)
            return JSONResponse(json.dumps(res))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/centroid/")
    def centroid(lang: str, words: List[str] = Query([])):
        try:
            words = filter_words(lang, words)
            res = models[lang].centroid(words)
            return JSONResponse(json.dumps(res.tolist()))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/analogy/")
    def analogy(lang: str, a: str, b: str, c: str, topn: int = 10):
        try:
            return JSONResponse(models[lang].analogy(a, b, c, topn))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/vector/")
    def vector(lang: str, word: str):
        try:
            res = models[lang].vector(word)
            return JSONResponse(json.dumps(res.tolist()))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/align/")
    def align(lang1: str, lang2: str, word: str, topn: int = 10):
        try:
            res = Embeddings.align(models[lang1], models[lang2], word=word, topn=topn)
            return JSONResponse(json.dumps(res))
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    @app.get("/vocabulary/")
    def vocabulary(lang: str):
        try:
            return JSONResponse(models[lang].L.index2word)
        except Exception as e:
            return JSONResponse({'error': str(e)}, status_code=500)

    return app


class EmbeddingsAPI:
    def __init__(self, host='http://localhost', port=1337, path=''):
        self.host = host
        self.port = port
        self.path = path
        self.baseurl = host + ':' + str(port) + path

    def n_similarity(self, ws1, ws2, lang='eng'):
        url = self.baseurl + '/n_similarity'
        content = self.get_response(url, {
            'ws1': ws1,
            'ws2': ws2,
            'lang': lang
        })
        return json.loads(content)

    def similarity(self, w1, w2, lang='eng'):
        url = self.baseurl + '/similarity'
        content = self.get_response(url, {
            'w1': w1,
            'w2': w2,
            'lang': lang
        })
        if content:
            return json.loads(content)
        return None

    def most_similar(self, positive, negative=[], topn=10, lang='eng'):
        url = self.baseurl + '/most_similar'
        content = self.get_response(url, {
            'positive': positive,
            'negative': negative,
            'topn': topn,
            'lang': lang
        })
        return json.loads(content)

    def vector(self, word, lang='eng'):
        url = self.baseurl + '/vector'
        content = self.get_response(url, {
            'word': word,
            'lang': lang
        })

        return json.loads(content)

    def model_word_set(self, lang='eng'):
        url = self.baseurl + '/model_word_set'
        content = self.get_response(url, {'lang': lang})
        return json.loads(content)

    def theme(self, words, lang='eng'):
        url = self.baseurl + '/theme'
        content = self.get_response(url, {'words': words, 'lang': lang})
        return json.loads(content)

    def neighbours(self, word, threshold=0.5, lang='eng'):
        url = self.baseurl + '/neighbours'
        content = self.get_response(url, {
            'word': word,
            'threshold': threshold,
            'lang': lang
        })

        return json.loads(content)

    def align(self, word, lang1='eng', lang2='fin'):
        url = self.baseurl + '/align'
        content = self.get_response(url, {
            'word': word,
            'lang1': lang1,
            'lang2': lang2
        })
        return json.loads(content)

    def centroid(self, words, lang='eng'):
        url = self.baseurl + '/centroid'
        content = self.get_response(url, {'words': words, 'lang': lang})
        return json.loads(content)

    def to_vector(self, tokens, lang='eng'):
        url = self.baseurl + '/to_vector'
        content = self.get_response(url, {'tokens': tokens, 'lang': lang})

        return json.loads(content)

    def analogy(self, a, b, c, topn=10, lang='eng'):
        url = self.baseurl + '/analogy'
        content = self.get_response(url, {
            'a': a,
            'b': b,
            'c': c,
            'topn': topn,
            'lang': lang
        })

        return json.loads(content)

    def vocabulary(self, lang='eng'):
        url = self.baseurl + '/vocabulary'
        content = self.get_response(url, {'lang': lang})
        return json.loads(content)

    def get_response(self, url, params={}):
        response = requests.get(url, params)
        if response.status_code != 200:
            content = json.loads(response.content)
            if 'error' in content:
                raise Exception(content['error'])
            else:
                raise Exception("Error... Response received: {}".format(content))
        return response.content


def test():
    api = EmbeddingsAPI()
    print(api.theme(words=['shoe', 'clothes'], lang='eng'))
    print(api.neighbours(word='hi', threshold=0.4, lang='eng'))
    print(api.analogy('man', 'king', 'woman', topn=10, lang='fin'))
    print(api.centroid(words=['hi', 'hello'], lang='eng'))
    print(api.to_vector(tokens='this is a great api !'.split(' '), lang='eng'))
    print(api.align(word='king', lang1='eng', lang2='fin'))
    print(api.similarity(w1='hi', w2='bye', lang='eng'))
    print(api.most_similar(positive=['hello', 'world'], negative=['king'], topn=10, lang='eng'))
    print(api.vector(word='king', lang='eng'))
    print(api.vocabulary(lang='eng'))
    # print(api.model_word_set(lang='eng'))


if __name__ == '__main__':
    test()



from semeval.common import download_path
import hickle as hkl
import numpy as np
import io
from sklearn import preprocessing
import pathlib


class Relatedness:
    def __init__(self, lang='eng'):

        rows_path = "{}{}-relatedness-{}.{}".format(download_path(), lang, 'rows', 'txt')
        cols_path = "{}{}-relatedness-{}.{}".format(download_path(), lang, 'cols', 'txt')
        matrix_path = "{}{}-relatedness-{}.{}".format(download_path(), lang, 'model', 'hkl')

        if not pathlib.Path(rows_path).is_file() or \
                not pathlib.Path(cols_path).is_file() or \
                not pathlib.Path(matrix_path).is_file():
            raise Exception(
                "Relatedness models are not download. Download them using `python -m semeval.download -m relatedness -l {}`"
                    .format(lang)
            )

        # get the rows and their indecies
        self.rows = Relatedness.load_termidx(rows_path)
        # get the columns and their indecies
        self.cols = Relatedness.load_termidx(cols_path)
        # load the matrix
        self.matrix = hkl.load(matrix_path)

        self.rev_cols = {i: r for r, i in self.cols.items()}
        sorted_cols = sorted(self.cols.items(), key=lambda k: k[1])
        self.sorted_cols = [r[0] for r in sorted_cols]

    @staticmethod
    def load_termidx(filename):
        with io.open(filename, 'r', encoding='utf-8') as f:
            rows = f.readlines()
            rows = [r.rstrip('\n').split('\t')[0] for r in rows]
            return dict([(w, i) for i, w in enumerate(rows)])

    def get_vector(self, word, normalize=True):
        try:
            i = self.rows[word]  # get the index of the word in the matrix
            row = self.matrix[i, :].A[0]  # get it's relatedness vector/row
            if normalize:
                row = preprocessing.normalize(row[:, np.newaxis], axis=0, norm='l1').ravel()
            return row
        except Exception as e:
            return None

    def get_rel(self, word, normalize=True, positive=True):
        try:
            i = self.rows[word]  # get the index of the word in the matrix
            row = self.matrix[i, :].A[0]  # get it's relatedness vector/row
            if normalize:
                row = preprocessing.normalize(row[:, np.newaxis], axis=0, norm='l1').ravel()
            row = zip(self.sorted_cols, row)  # word: relatedness_score
            if positive:
                row = [r for r in row if r[1] > 0]  # remove non-related words
            row = dict(row)
            return row
        except Exception as e:
            return None

    def get_sorted_rel(self, word, normalize=True, positive=True, k=0):
        row = self.get_rel(word, normalize, positive)
        if row:
            row = sorted(row.items(), key=lambda k: k[1], reverse=True)
            if k > 0 and k < len(row):
                row = row[:k]
        return row

    def interpret(self, tenor, vehicle):
        """
        The "Combined rank" interpretation method as presented in
        Xiao, P., Alnajjar, K., Granroth-Wilding, M., Agres, K., & Toivonen, H. (2016). Meta4meaning: Automatic Metaphor
         Interpretation Using Corpus-Derived Word Associations. In Proceedings of The Seventh International Conference
         on Computational Creativity (pp. 230-237). Sony CSL Paris.
        """
        tv = self.get_vector(tenor)
        vv = self.get_vector(vehicle)

        if tv is None or vv is None:
            return []

        features = np.union1d(np.where(tv > 0), np.where(vv > 0))  # all non-zero features
        shared_features = np.intersect1d(np.where(tv > 0), np.where(vv > 0))  # shared features

        # consider only concrete features? add filter here (e.g., certain POS tags)

        mv = np.zeros(tv.shape)
        mv[shared_features] = tv[shared_features] * vv[shared_features]  # multiplication
        mv = mv[features]

        odv = np.zeros(tv.shape)
        odv[features] = float('-inf')
        odv[shared_features] = vv[shared_features] - tv[shared_features]  # overlap difference
        odv = odv[features]

        # sort them and convert back into ordered features
        mvf = np.argsort(-mv)
        odvf = np.argsort(-odv)

        mvf_l = mvf.tolist()
        odvf_l = odvf.tolist()

        # ranked interpretations
        interpretations = [(self.rev_cols[features[f]], min([mvf_l.index(f), odvf_l.index(f)])) for f in
                           range(len(features))]
        return list(sorted(interpretations, key=lambda k: k[1]))

    def metaphoricity(self, tenor, vehicle, expression, k=0, normalize=True):
        """
        Metaphoricity scores as described in Alnajjar, K., & Toivonen, H. (2020). Computational Generation of Slogans.
         Natural Language Engineering. https://doi.org/10.1017/S1351324920000236
        """
        tv = self.get_sorted_rel(tenor, normalize=normalize)
        vv = self.get_sorted_rel(vehicle, normalize=normalize)

        if not tv or not vv:
            return 0.0

        if k == 0:
            pass
        elif k > 0 and k < len(self.rows):
            tv = tv[:k]
            vv = vv[:k]
        else:
            raise Exception("k must be positive and less than %s" % len(self.rows))

        tv = dict(tv)
        vv = dict(vv)

        t_relatedness = np.max([tv.get(t, 0.0) for t in expression])  # relatedness to tenor
        v_relatedness = np.max([vv.get(t, 0.0) for t in expression])  # relatedness to vehicle
        tv_score = t_relatedness * v_relatedness

        vt_diff = np.max([vv.get(t, 0.0) - tv.get(t, 0.0) for t in expression])  # vehicle tenor difference

        return tv_score, vt_diff, np.mean([tv_score, vt_diff]) if tv_score > 0 and vt_diff > 0 else 0.0


def main():
    from pprint import pprint

    relatedness_m = Relatedness(lang='eng')

    pprint(relatedness_m.get_sorted_rel('car')[:5])
    print()

    pprint(relatedness_m.interpret('alcohol', 'crutch')[:10])
    print()
    pprint(relatedness_m.interpret('cloud', 'cotton')[:10])
    print()

    # 0 to select all
    pprint(relatedness_m.metaphoricity('life', 'journey', ['I', 'walked', 'in', 'his', 'sorrows', '.'], 300))


if __name__ == '__main__':
    main()
